@@ -60,17 +60,6 @@ to the *hicsv*-formatted files.
 It, therefore, does not have (or aim for) any advanced functionalities. 
 Please use ``numpy`` and/or other useful packages for anything beyond I/O. 
 
-Note:
-    Some users may find the following behavior of this package unexpected. 
-
-    1. The parser deletes the spaces at the beginning and at the end of each cell. 
-       This is true even if the cell content is double-quoted. 
-       For example, the two cells
-
-            "spam", "  spam  "
-       
-       results in identical strings. 
-
 Example:
     To read a hicsv-formatted text file,
 
@@ -105,15 +94,26 @@ Example:
     so adding an array with different length will cause error. 
     Then you can write into a hicsv-formatted text file like::
 
-        >>> out.save("new hicsv file.txt")
+        >>> out.save("new hicsv file.txt", encoding="utf-8")
+
+Note:
+    Some users may find the following behavior of this package unexpected. 
+
+    1. The parser deletes the spaces at the beginning and at the end of each cell. 
+       This is true even if the cell content is double-quoted. 
+       For example, the two cells
+
+            "spam", "  spam  "
+       
+       results in identical strings. 
 """
 
-from typing import List, IO
+from typing import List, IO, Any
 import json, csv, io
 
 import numpy as np
 
-__version__ = "0.0.0"
+__version__ = "1.0.0"
 HICSV_VERSION = "20220812"
 
 HEADER_STARTSWITH: str = "#"
@@ -123,12 +123,12 @@ class hicsv(object):
     """Container/writer class for a hicsv file. 
 
     Args:
-        fp: 
-            File-like object or path to the file. 
+        fp: File-like object or path to the file. 
             If an empty string, an empty object will be created. 
             If a string or a file-like object is given, 
             the file will be loaded as the object. 
-            Defaults to "" (empty string). 
+            Defaults to ``""`` (empty string) which creates an empty object. 
+        **kwargs: Arguments of built-in function ``open()``. (`New in 1.0.0.`)
 
     Attributes:
         keys: column keys. a list of strings. 
@@ -154,7 +154,7 @@ class hicsv(object):
 
     """
 
-    def __init__(self, fp: str|IO = "") -> None:
+    def __init__(self, fp: str|IO = "", **kwargs: Any) -> None:
         """initializer
         """
         self.keys: List[str]        = []
@@ -163,7 +163,7 @@ class hicsv(object):
 
         if fp:
             if isinstance(fp, str):
-                with open(fp, "r") as f:
+                with open(fp, **kwargs) as f:
                     self._fromfile(f)
             else:
                 self._fromfile(fp)
@@ -201,6 +201,8 @@ class hicsv(object):
         keys: List[str] = []
         rows_str: List[List[str]] = []
         cols_str: List[List[str]] = []
+        num_rows: int = 0
+        num_cols: int = 0
 
         lines = fp.readlines()
 
@@ -267,9 +269,9 @@ class hicsv(object):
             if row:
                 rows_str.append([e.strip() for e in row]) # spaces are omitted
 
-        length = len(rows_str)
-
-        cols_str = [*zip(*rows_str)]
+        num_rows = len(rows_str)
+        num_cols = len(keys)
+        cols_str = [[rows_str[i][j] for i in range(num_rows)] for j in range(num_cols)] # same as list(zip(*row_str))
 
         # detect type and store arrays
         for key, col_str in zip(keys, cols_str):
@@ -298,15 +300,15 @@ class hicsv(object):
         Returns:
             column index. 
         """
-        try:
-            return self.keys.index(col)
-        except ValueError:
-            if isinstance(col, int):
-                if col < len(self.keys):
-                    return col
-                else:
-                    raise IndexError("column index out of range: max. {0} but {1} given".format(len(self.keys), col))
+        if isinstance(col, int):
+            if col < len(self.keys):
+                return col
             else:
+                raise IndexError("column index out of range: max. {0} but {1} given".format(len(self.keys), col))
+        elif isinstance(col, str):
+            try:
+                return self.keys.index(col)
+            except ValueError:
                 raise KeyError("column key not found: '{0}'".format(col))
 
     def _get_column_as_array(self, col: int|str) -> np.ndarray:
@@ -338,8 +340,7 @@ class hicsv(object):
         if len(cols) == 1:
             return self._get_column_as_array(cols[0])
         else:
-            for col in cols:
-                return [self._get_column_as_array(col) for col in cols]
+            return [self._get_column_as_array(col) for col in cols]
 
     def ga(self, *cols: int|str) -> np.ndarray | List[np.ndarray]:
         """Alias for `get_columns_as_arrays()`. 
@@ -393,7 +394,7 @@ class hicsv(object):
         elif not (np.issubdtype(arr.dtype, np.integer) or 
                   np.issubdtype(arr.dtype, np.floating) or 
                   np.issubdtype(arr.dtype, np.str_)):
-            raise TypeError("expected numpy array with the sub-dtype of numpy.integer, numpy.floating, or numpy.str_: got".format(arr.dtype))
+            raise TypeError("expected numpy array with the sub-dtype of numpy.integer, numpy.floating, or numpy.str_: got <{0}>".format(arr.dtype))
         elif (not self.keys) and pos==0: # first column!
             ok = True
         elif key in self.keys:
@@ -545,7 +546,7 @@ class hicsv(object):
         
         fp.write("\r\n".join([DELIMITER.join(row) for row in zip(*scols)]))
 
-    def save(self, fp: str|IO, prettify: bool = True, add_version_info: bool = True) -> None:
+    def save(self, fp: str|IO, prettify: bool = True, add_version_info: bool = True, **kwargs: Any) -> None:
         """Saves the object content to a hicsv-formatted text file. 
         
         Args:
@@ -553,6 +554,8 @@ class hicsv(object):
             prettify: True to turn on pretty formatting. 
             add_version_info: If True, automatically adds the current versions
                 of the module and the hicsv format specification. 
+            **kwargs: Arguments of built-in ``open()``. Used only when ``fp`` is a string. 
+                The ``mode`` parameter is set to "w" by default. (`New in 1.0.0.`)
         
         Example:
             >>> d = hicsv.hicsv()
@@ -567,7 +570,10 @@ class hicsv(object):
             >>>     d.save(fp)
         """
         if isinstance(fp, str):
-            with open(fp, "w") as f:
+            _kwargs: dict = {}
+            _kwargs["mode"] = "w"
+            _kwargs.update(**kwargs)
+            with open(fp, "w", **_kwargs) as f:
                 return self._save(f, prettify, add_version_info)
         else:
             return self._save(fp, prettify, add_version_info)
@@ -582,7 +588,7 @@ class hicsv(object):
 
 
 
-def txt2hicsv(fp: IO|str, sep: str = ",", ignore_lines: List[int] = [], key_line: int|None = None, keys: List[str] = []) -> hicsv:
+def txt2hicsv(fp: IO|str, sep: str = ",", ignore_lines: List[int] = [], key_line: int|None = None, keys: List[str] = [], **kwargs: Any) -> hicsv:
     '''Reads a generic delimited text file and converts it into a hicsv object. 
 
     This is a convenient function to import simple text data file into a hicsv object. 
@@ -597,29 +603,33 @@ def txt2hicsv(fp: IO|str, sep: str = ",", ignore_lines: List[int] = [], key_line
         keys: List of column keys. It's useful if you know the table structure exactly. 
             This must match the number of columns in the input file. 
             When `keys` is specified, `key_line` will be ignored. 
+        **kwargs: Arguments of built-in ``open()``. Used only when ``fp`` is a string. (`New in 1.0.0.`)
     '''
 
-    lines: List(str) = []
+    lines: List[str] = []
     lines_table: List[str] = []
-    rows_str: List(str) = []
-    _ignore_lines: List(int) = ignore_lines
+    rows_str: List[List[str]] = []
+    _ignore_lines: List[int] = ignore_lines
     i: int = 0
-    _keys: List(str) = []
+    _keys: List[str] = []
     length: int = 0
 
     if isinstance(fp, str):
-        with open(fp, "r") as f:
+        with open(fp, **kwargs) as f:
             lines = f.readlines()
     else:
         lines = fp.readlines()
 
-    # key line must also be ignored
+    # if keys are provided, use that. 
     if keys:
         _keys = keys
-    elif key_line != None:
+    # if not provided and the key_line is provided (as int), 
+    # try to read the keys from that line
+    elif isinstance(key_line, int):
         for row in csv.reader([lines[key_line], ], delimiter=sep): # just one line
             _keys = [k.strip(" ") for k in row]
 
+        # this line should also be ignored
         _ignore_lines.append(key_line)
 
     # ignore the specified lines
